#ifndef ARBUTILS_UNIQUEPTR_HPP
#define ARBUTILS_UNIQUEPTR_HPP

#include "../Assert.hpp"

namespace ArbUt {
    /// @brief A unique pointer is used to indicate a pointer that is owned by its holder, and will be deleted when its
    /// owner is deleted. As with all Arbutils pointers, this cannot be assigned null. Use an OptionalUniquePtr for pointers that can be null.
    /// @details A unique pointer is used to indicate a pointer that is owned by an object, and that needs to be deleted
    /// when its owner is deleted.
    template <class T> class UniquePtr {
    private:
        T* _raw;

    public:
        /// @brief Initialise a UniquePtr with a specific raw pointer.
        inline UniquePtr<T>(__attribute__((nonnull)) T* ptr) : _raw(ptr) { AssertNotNull(ptr); };
        /// @brief Initialise a UniquePtr from a copy.
        inline UniquePtr<T>(const UniquePtr<T>& other) : _raw(other._raw){};
        /// @brief Initialise a UniquePtr with a std unique_ptr.
        inline UniquePtr<T>(const std::unique_ptr<T>& other) : _raw(other.get()){};

        ~UniquePtr() noexcept{
            delete _raw;
        }

        /// @brief Copy operator.
        inline UniquePtr<T>& operator=(const UniquePtr<T>& rhs) {
            if (this == &rhs)
                return *this;
            _raw = rhs._raw;
            return *this;
        }

        inline UniquePtr<T>& operator=(__attribute__((nonnull)) T* rhs) {
            if (_raw == &rhs)
                return *this;
            AssertNotNull(rhs);
            _raw = rhs;
            return *this;
        }

        /// @brief Operator for access into underlying pointer.
        /// @warning Note that this asserts that the underlying pointer is not null first, to prevent segfaults.
        inline T* operator->() const noexcept { return _raw; }

        /// @brief Get the raw underlying pointer.
        inline T* GetRaw() const noexcept { return _raw; }

        /// @brief Check equality of two UniquePtr objects
        inline bool operator==(const UniquePtr& rhs) const noexcept { return _raw == rhs._raw; }
        /// @brief Check equality of pointers
        inline bool operator==(T* rhs) const noexcept { return _raw == rhs; }
        /// @brief Check equality of two UniquePtr objects
        inline bool operator!=(const UniquePtr& rhs) const noexcept { return _raw != rhs._raw; }
        /// @brief Check equality of pointers
        inline bool operator!=(T* rhs) const noexcept { return _raw == rhs; }
    };
}

namespace std {
    /// @brief Helper class for allowing hashing of UniquePtr.
    template <class T> struct hash<ArbUt::UniquePtr<T>> {
        /// @brief Returns a hash of for a borrowed Pointer. Effectively just the raw memory address.
        /// @param k A borrowed pointer.
        /// @return The hash of the borrowed pointer.
        std::size_t operator()(const ArbUt::UniquePtr<T>& k) const { return (size_t)k.GetRaw(); }
    };
}

#endif // ARBUTILS_UNIQUEPTR_HPP
